#ifndef __LIBSANGOMA_FREETDM_H__
#define __LIBSANGOMA_FREETDM_H__

#include "freetdm.h"
#include "ftdm_os.h"
#include "libteletone.h"
#include "ftdm_buffer.h"
#include "libsangoma_wp_config.h"

#define FTDM_MAX_SPANS_INTERFACE 128
#define MAX_QUEUE_SIZE 1024

//FIO_SIGNAL_CB_FUNCTION(on_freetdm_event_signal);

#define SNG_FTDM_IS_DCHAN(ftdmchan) (ftdm_channel_get_type(ftdmchan) == FTDM_CHAN_TYPE_DQ921 || ftdm_channel_get_type(ftdmchan)  == FTDM_CHAN_TYPE_DQ931)

typedef struct sangoma_freetdm_init {
	ftdm_queue_t *event_msg_queue;
	uint32_t event_msg_queue_sz;
	uint32_t logging_level;
	uint32_t crash_policy_level;
	ftdm_logger_t logger;
} sangoma_freetdm_init_t;

/*!
  \ in order to provide user flexibility to configure
  \FreeTDM span for different modules

*/
typedef struct span_config {
	union {
		struct {
			char mixaudio_mode[10];
			char interface_name[10];
			char peer_span[10];
			char interface_type[10];
			char switch_type[10];
			char debug[10];
			char call_info_indication[10];
		} isdn;

		struct {
			char debug[100];
		} gsm;

		struct {
			int reserved;
		} fxo;

		struct {
			char mixaudio_mode[10];
			char peer_span[10];
			char switch_type[10];
			char cicbase[10];
			char is_voice[10];
			char voice_span[20];
		} isup;

		struct {
			char span_name[10];
			char cfgprofile[15];
			char p_name[10][10];
			char p_value[10][10];
			ftdm_conf_node_t *ss7confnode;
		} mtp2_api;
	} config;
} span_config_t;

/*!
  \fn int sangoma_freetdm_init(sangoma_freetdm_init_t *ftdm_init)
  \brief Initialize FreeTDM Framework. Must run before any function.
  \param ftdm_init FreeTDM initi config structure
  \return File Descriptor: Non Zero is error, 0  is success

  Main FreeTDM initializatino function.
  Must be the first functino to be called before any other FreeTDM call.

*/
int sangoma_freetdm_init(sangoma_freetdm_init_t * ftdm_init);

/*!
  \fn int sangoma_freetdm_destroy(void);
  \brief Stop the FreeTDM Framework. All Spans will be deallocated.
  \param void
  \return File Descriptor: Non Zero is error, 0  is success

  Stop the FreeTDM Framework.
  All spans will be deallocated.  
  Usually called before the application exits.
  Must be the first function to be called before any other FreeTDM call.

*/
int sangoma_freetdm_destroy(void);

/*!
  \fn int sangoma_freetdm_create_span(sangoma_port_obj_t port, ftdm_span_t ** span, int operating_mode);
  \brief Create FreeTDM Span. User hw port object for span cfg info.
  \param port Sangoma hw prot obj. 
  \param span Created span will be stored in this pointer. Span is an opaque object.
  \param operating_mode will contain the mode on which span needs to be operated like ANALOG/ISUP_SIG/ISUP_VOICE/ISDN/MTP2_API_LSL/MTP2_API_HSL
  \return File Descriptor: Non Zero is error, 0  is success

  Create FreeTDM Span in FreeTDM framework.
  The Sangoma port object will be used to extract hw configuration necessary
  to create the span.

  The FreeTDM span opaque object will be stored in the span argument.
  At this point span is only half configured. 
  Next step is to call span configuration function.
*/
int sangoma_freetdm_create_span(sangoma_port_obj_t port, ftdm_span_t ** span, int operating_mode);

/*!
  \fn int sangoma_freetdm_span_config_and_start(sangoma_port_obj_t port, ftdm_span_t *span);
  \brief Configure FreeTDM span based on sangoma port object configuration options.
  \param port Sangoma hw prot obj. 
  \param span FreeTDM span object
  \param operating_mode will contain the mode on which span needs to be configured and started like ANALOG/ISUP_SIG/ISUP_VOICE/ISDN/MTP2_API_LSL/MTP2_API_HSL
  \return File Descriptor: Non Zero is error, 0  is success

  This function will configure the FreeTDM span using information from sangoma port obj.
  Once the span is fully configured, the span will be started.

  Once the FreeTDM span starts it will
    Initialize and start signaling on port dchan
    Wait for dchan events.
    One signaling event (such as call start) FreeTDM pass the event structure
    to user application via callback function.

  At this point application is able to call FreeTDM Signaling and Media Functions.
*/
int sangoma_freetdm_span_config_and_start(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t * span_conf, int operating_mode);

int sangoma_freetdm_send_sms(ftdm_span_t * span, char *called, char *text);
int sangoma_freetdm_call_req(ftdm_span_t * span, char *called, char *calling);
int sangoma_freetdm_send_at(ftdm_span_t * span, char *command);
int sangoma_freetdm_send_debug(ftdm_span_t * span, char *command);

int sangoma_freetdm_gsm_span_status(ftdm_span_t * span);
int sangoma_freetdm_channel_dump(ftdm_span_t * span, int);
int sangoma_freetdm_channel_list(ftdm_span_t * span, int);

/* For ISDN TAP calls statistics */
int sangoma_freetdm_isdn_tap_statistics(void);

/* Logger Functions */
 /*!
  \Declare SDK log level depending up on FREETDM_LOG_LEVEL as shown below:
	FTDM_LOG_LEVEL_DEBUG 7
	FTDM_LOG_LEVEL_INFO 6
	FTDM_LOG_LEVEL_NOTICE 5
	FTDM_LOG_LEVEL_WARNING 4
	FTDM_LOG_LEVEL_ERROR 3
	FTDM_LOG_LEVEL_CRIT 2
	FTDM_LOG_LEVEL_ALERT 1
	FTDM_LOG_LEVEL_EMERG 0
*/
static const char *SANGOMA_FTDM_LEVEL_NAMES[9] = {
	"EMERG",
	"ALERT",
	"CRIT",
	"ERROR",
	"WARNING",
	"NOTICE",
	"INFO",
	"DEBUG",
	 NULL
};

/*!
  \fn const char* sangoma_freetdm_cut_path(const char *in);
  \brief Take out the file name from where loggnig is being called.
  \param file_path this is the full file path form where logging is being called
  \return constant character pointer if file name is present else NULL
*/
const char* sangoma_freetdm_cut_path(const char *file_path);

/*!
  \fn void sangoma_freetdm_logger(const char *file, const char *func, int line, int level, const char *fmt, ...);
  \brief print the log on to the standard output device i.e. monitor
  \param file File path
  \param func Function name
  \param line Line number
  \param level Log Level
  \param fmt Log/message that needs to be printed
  \return prints the log and return nothing
*/
void sangoma_freetdm_logger(const char *file, const char *func, int line, int level, const char *fmt, ...);

/*!
  \fn void sangoma_freetdm_exec_io_command(const char *cmd, ftdm_channel_t *fchan)
  \brief cmd = enable/disable/flush or get statistics 
  \param fchan = FTDM channel from where stats operation(cmd) will performed 
  \return prints the stats or perform enable/disable/flush operation/cmd and return nothing
*/
void sangoma_freetdm_exec_io_command(const char *cmd, ftdm_channel_t *fchan);

/*!
  \fn void sangoma_freetdm_enable_io_stats(ftdm_span_t *span)
  \brief span = execute stats cmd all channels on this span 
*/
void sangoma_freetdm_exec_io_stats(const char* cmd, ftdm_span_t * span);

/*!
  \fn void sangoma_freetdm_set_sig_rx_queuesize(ftdm_channel_t *fchan, int rxsize);
  \brief Set the D channel Wanpipe Queue size
*/
void sangoma_freetdm_set_chan_sig_rx_queuesize(ftdm_channel_t *fchan, int rxsize);


/*!
  \fn void sangoma_freetdm_set_span_sig_queuesize(ftdm_span_t *span, int rxsize);
  \brief Set the D channel Wanpipe Queue size for all channels belongs to span
*/
void sangoma_freetdm_set_span_sig_queuesize(ftdm_span_t *span, int rxsize);




#endif
